@@ -21,10 +21,12 @@ from keras_tuner.engine.hyperparameters import HyperParameter
 from keras_tuner.engine.hyperparameters import HyperParameters
 from keras_tuner.engine.logger import CloudLogger
 from keras_tuner.engine.logger import Logger
-from keras_tuner.engine.oracle import Objective
+from keras_tuner.engine.objective import Objective
 from keras_tuner.engine.oracle import Oracle
+from keras_tuner.engine.oracle import synchronized
 from keras_tuner.engine.tuner import Tuner
 from keras_tuner.tuners import BayesianOptimization
+from keras_tuner.tuners import GridSearch
 from keras_tuner.tuners import Hyperband
 from keras_tuner.tuners import RandomSearch
 from keras_tuner.tuners import SklearnTuner
@@ -32,4 +34,4 @@ from keras_tuner.utils import check_tf_version
 
 check_tf_version()
 
-__version__ = "1.1.0"
+__version__ = "1.2.0"
