@@ -1,4 +1,4 @@
-# Copyright 2020 Google LLC. All Rights Reserved.
+# Copyright 2019 The KerasTuner Authors
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -167,8 +167,8 @@ class HyperImageAugment(hypermodel.HyperModel):
                 and isinstance(augment_layers_max, int)
             ):
                 raise ValueError(
-                    "Keyword argument `augment_layers` must be int,"
-                    "but received {}. ".format(augment_layers)
+                    "Keyword argument `augment_layers` must be "
+                    f"int,but received {augment_layers}."
                 )
 
             self.augment_layers_min = augment_layers_min
@@ -178,7 +178,7 @@ class HyperImageAugment(hypermodel.HyperModel):
             # `randaug_count` is set to 0.
             self.model_name = "image_augment"
 
-        super(HyperImageAugment, self).__init__(**kwargs)
+        super().__init__(**kwargs)
 
     def build(self, hp):
         if self.input_tensor is not None:
@@ -259,9 +259,7 @@ class HyperImageAugment(hypermodel.HyperModel):
             transform_factor_max = transform_params[1]
             if len(transform_params) > 2:
                 raise ValueError(
-                    "Length of keyword argument {} must not exceed 2.".format(
-                        transform_name
-                    )
+                    f"Length of keyword argument {transform_name} must not exceed 2."
                 )
         except TypeError:
             transform_factor_min = 0
@@ -272,8 +270,8 @@ class HyperImageAugment(hypermodel.HyperModel):
             and isinstance(transform_factor_min, (int, float))
         ):
             raise ValueError(
-                "Keyword argument {} must be int or float, "
-                "but received {}. ".format(transform_name, transform_params)
+                f"Keyword argument {transform_name} must be int "
+                f"or float, but received {transform_params}."
             )
 
         self.transforms.append(
