@@ -48,12 +48,12 @@ class OracleServicer(service_pb2_grpc.OracleServicer):
         status = self.oracle.update_trial(
             request.trial_id, request.metrics, step=request.step
         )
-        status_proto = trial_module._convert_trial_status_to_proto(status)
+        status_proto = trial_module.TrialStatus.to_proto(status)
         return service_pb2.UpdateTrialResponse(status=status_proto)
 
     def EndTrial(self, request, context):
-        status = trial_module._convert_trial_status_to_str(request.status)
-        self.oracle.end_trial(request.trial_id, status)
+        trial = trial_module.Trial.from_proto(request.trial)
+        self.oracle.end_trial(trial)
         return service_pb2.EndTrialResponse()
 
     def GetTrial(self, request, context):
@@ -74,7 +74,7 @@ def start_server(oracle):
     server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
     oracle_servicer = OracleServicer(oracle)
     service_pb2_grpc.add_OracleServicer_to_server(oracle_servicer, server)
-    server.add_insecure_port("{}:{}".format(ip_addr, port))
+    server.add_insecure_port(f"{ip_addr}:{port}")
     server.start()
     while True:
         # The server does not block otherwise.
@@ -85,8 +85,10 @@ def start_server(oracle):
                 time.sleep(10)
 
             print(
-                "Oracle server on chief is exiting in 10s."
+                "Oracle server on chief is exiting in 40s."
                 "The chief will go on with post-search code."
             )
+            # Wait for another 30s for all the subsequent calls from workers to come
+            time.sleep(30)
             server.stop(10)
             break
