@@ -1,4 +1,4 @@
-# Copyright 2019 The Keras Tuner Authors
+# Copyright 2019 The KerasTuner Authors
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,19 +13,18 @@
 # limitations under the License.
 "HyperParameters logic."
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
 
 import abc
 
 import six
+from tensorflow import keras
 
-from ..protos import keras_tuner_pb2
+from keras_tuner import utils
+from keras_tuner.protos import keras_tuner_pb2
 
 
 @six.add_metaclass(abc.ABCMeta)
-class Condition(object):
+class Condition:
     """Abstract condition for a conditional hyperparameter.
 
     Subclasses of this object can be passed to a `HyperParameter` to specify
@@ -61,7 +60,7 @@ class Condition(object):
         return cls(**config)  # pytype: disable=not-instantiable
 
     @classmethod
-    def from_proto(self, proto):
+    def from_proto(cls, proto):
         kind = proto.WhichOneof("kind")
         if kind == "parent":
             parent = getattr(proto, kind)
@@ -69,7 +68,7 @@ class Condition(object):
             values = parent.values
             values = [getattr(v, v.WhichOneof("kind")) for v in values]
             return Parent(name=name, values=values)
-        raise ValueError("Unrecognized condition of type: {}".format(kind))
+        raise ValueError(f"Unrecognized condition of type: {kind}")
 
 
 class Parent(Condition):
@@ -98,7 +97,7 @@ class Parent(Condition):
         self.name = name
 
         # Standardize on str, int, float, bool.
-        values = _to_list(values)
+        values = utils.to_list(values)
         first_val = values[0]
         if isinstance(first_val, six.string_types):
             values = [str(v) for v in values]
@@ -138,9 +137,16 @@ class Parent(Condition):
         )
 
 
-def _to_list(values):
-    if isinstance(values, list):
-        return values
-    if isinstance(values, tuple):
-        return list(values)
-    return [values]
+OBJECTS = (
+    Condition,
+    Parent,
+)
+ALL_CLASSES = {cls.__name__: cls for cls in OBJECTS}
+
+
+def deserialize(config):
+    return keras.utils.deserialize_keras_object(config, module_objects=ALL_CLASSES)
+
+
+def serialize(obj):
+    return keras.utils.serialize_keras_object(obj)
