@@ -18,19 +18,15 @@ import contextlib
 import copy
 import gc
 import os
-import traceback
 
 import numpy as np
-import tensorflow as tf
 from tensorboard.plugins.hparams import api as hparams_api
 from tensorflow import keras
 
-from keras_tuner import config as config_module
+from keras_tuner import errors
 from keras_tuner.engine import base_tuner
 from keras_tuner.engine import tuner_utils
 
-MAX_FAIL_STREAK = 5
-
 
 class Tuner(base_tuner.BaseTuner):
     """Tuner class for Keras models.
@@ -39,6 +35,10 @@ class Tuner(base_tuner.BaseTuner):
     the building, training, evaluation and saving of the Keras models. New
     tuners can be created by subclassing the class.
 
+    All Keras related logics are in `Tuner.run_trial()` and its subroutines.
+    When subclassing `Tuner`, if not calling `super().run_trial()`, it can tune
+    anything.
+
     Args:
         oracle: Instance of `Oracle` class.
         hypermodel: Instance of `HyperModel` class (or callable that takes
@@ -47,10 +47,10 @@ class Tuner(base_tuner.BaseTuner):
             `self.hypermodel`.
         max_model_size: Integer, maximum number of scalars in the parameters of
             a model. Models larger than this are rejected.
-        optimizer: Optional `Optimizer` instance.  May be used to override the
-            `optimizer` argument in the `compile` step for the models. If the
-            hypermodel does not compile the models it generates, then this
-            argument must be specified.
+        optimizer: Optional optimizer. It is used to override the `optimizer`
+            argument in the `compile` step for the models. If the hypermodel
+            does not compile the models it generates, then this argument must be
+            specified.
         loss: Optional loss. May be used to override the `loss` argument in the
             `compile` step for the models. If the hypermodel does not compile
             the models it generates, then this argument must be specified.
@@ -108,7 +108,7 @@ class Tuner(base_tuner.BaseTuner):
                 "using a `HyperModel` instance."
             )
 
-        super(Tuner, self).__init__(
+        super().__init__(
             oracle=oracle,
             hypermodel=hypermodel,
             directory=directory,
@@ -117,18 +117,12 @@ class Tuner(base_tuner.BaseTuner):
             overwrite=overwrite,
         )
 
-        if isinstance(oracle.objective, list) and len(oracle.objective) > 1:
-            raise ValueError(
-                "Multi-objective is not supported, found: {}".format(
-                    oracle.objective
-                )
-            )
-
         self.max_model_size = max_model_size
         self.optimizer = optimizer
         self.loss = loss
         self.metrics = metrics
         self.distribution_strategy = distribution_strategy
+        self.executions_per_trial = executions_per_trial
 
         # Support multi-worker distribution strategies w/ distributed tuning.
         # Only the chief worker in each cluster should report results.
@@ -142,17 +136,8 @@ class Tuner(base_tuner.BaseTuner):
                 self.distribution_strategy.extended.should_checkpoint
             )
 
-        # Save only the last N checkpoints.
-        self._save_n_checkpoints = 10
-
         self.tuner_id = tuner_id or self.tuner_id
 
-        self.executions_per_trial = executions_per_trial
-        # This is the `step` that will be reported to the Oracle at the end
-        # of the Trial. Since intermediate results are not used, this is set
-        # to 0.
-        self._reported_step = 0
-
     def _build_hypermodel(self, hp):
         with maybe_distribute(self.distribution_strategy):
             model = self.hypermodel.build(hp)
@@ -160,39 +145,24 @@ class Tuner(base_tuner.BaseTuner):
             return model
 
     def _try_build(self, hp):
-        for i in range(MAX_FAIL_STREAK + 1):
-            # clean-up TF graph from previously stored (defunct) graph
-            keras.backend.clear_session()
-            gc.collect()
+        # clean-up TF graph from previously stored (defunct) graph
+        keras.backend.clear_session()
+        gc.collect()
 
-            # Build a model, allowing max_fail_streak failed attempts.
-            try:
-                model = self._build_hypermodel(hp)
-            except:
-                if config_module.DEBUG:
-                    traceback.print_exc()
-
-                print("Invalid model %s/%s" % (i, MAX_FAIL_STREAK))
-
-                if i == MAX_FAIL_STREAK:
-                    raise RuntimeError("Too many failed attempts to build model.")
-                continue
-
-            # Stop if `build()` does not return a valid model.
-            if not isinstance(model, keras.models.Model):
-                raise RuntimeError(
-                    "Model-building function did not return "
-                    "a valid Keras Model instance, found {}".format(model)
-                )
-
-            # Check model size.
-            size = maybe_compute_model_size(model)
-            if self.max_model_size and size > self.max_model_size:
-                print("Oversized model: {} parameters -- skipping".format(size))
-                if i == MAX_FAIL_STREAK:
-                    raise RuntimeError("Too many consecutive oversized models.")
-                continue
-            break
+        model = self._build_hypermodel(hp)
+        # Stop if `build()` does not return a valid model.
+        if not isinstance(model, keras.models.Model):
+            raise errors.FatalTypeError(
+                "Expected the model-building function, or HyperModel.build() to "
+                "return a valid Keras Model instance. "
+                f"Received: {model} of type {type(model)}."
+            )
+        # Check model size.
+        size = maybe_compute_model_size(model)
+        if self.max_model_size and size > self.max_model_size:
+            raise errors.FailedTrialError(
+                f"Oversized model: {size} parameters. Skip model."
+            )
         return model
 
     def _override_compile_args(self, model):
@@ -206,7 +176,14 @@ class Tuner(base_tuner.BaseTuner):
                 if self.loss:
                     compile_kwargs["loss"] = self.loss
                 if self.optimizer:
-                    compile_kwargs["optimizer"] = self.optimizer
+                    optimizer = (
+                        self.optimizer
+                        if isinstance(self.optimizer, str)
+                        else keras.optimizers.deserialize(
+                            keras.optimizers.serialize(self.optimizer)
+                        )
+                    )
+                    compile_kwargs["optimizer"] = optimizer
                 if self.metrics:
                     compile_kwargs["metrics"] = self.metrics
                 model.compile(**compile_kwargs)
@@ -231,7 +208,11 @@ class Tuner(base_tuner.BaseTuner):
         """
         hp = trial.hyperparameters
         model = self._try_build(hp)
-        return self.hypermodel.fit(hp, model, *args, **kwargs)
+        results = self.hypermodel.fit(hp, model, *args, **kwargs)
+        tuner_utils.validate_trial_results(
+            results, self.oracle.objective, "HyperModel.fit()"
+        )
+        return results
 
     def run_trial(self, trial, *args, **kwargs):
         """Evaluates a set of hyperparameter values.
@@ -282,12 +263,11 @@ class Tuner(base_tuner.BaseTuner):
             of results of any of the types above. The final objective value is
             the average of the results in the list.
         """
-        model_checkpoint = keras.callbacks.ModelCheckpoint(
-            filepath=self._get_checkpoint_fname(trial.trial_id, self._reported_step),
-            monitor=self.oracle.objective.name,
-            mode=self.oracle.objective.direction,
-            save_best_only=True,
-            save_weights_only=True,
+        # Not using `ModelCheckpoint` to support MultiObjective.
+        # It can only track one of the metrics to save the best model.
+        model_checkpoint = tuner_utils.SaveBestEpoch(
+            objective=self.oracle.objective,
+            filepath=self._get_checkpoint_fname(trial.trial_id),
         )
         original_callbacks = kwargs.pop("callbacks", [])
 
@@ -303,32 +283,15 @@ class Tuner(base_tuner.BaseTuner):
             copied_kwargs["callbacks"] = callbacks
             obj_value = self._build_and_fit_model(trial, *args, **copied_kwargs)
 
-            # objective left unspecified,
-            # and objective value is not a single float.
-            if (
-                not isinstance(obj_value, (int, float))
-                and self.oracle.objective.name == "default_objective"
-            ):
-                raise TypeError(
-                    "Expect the return value of `Tuner.run_trial()` or "
-                    "`HyperModel.fit()` to be a single float when "
-                    "`objective` is left unspecified. Recevied return value: "
-                    f"{obj_value}."
-                )
-
             histories.append(obj_value)
         return histories
 
     def load_model(self, trial):
         model = self._try_build(trial.hyperparameters)
-        # Reload best checkpoint. The Oracle scores the Trial and also
-        # indicates at what epoch the best value of the objective was
-        # obtained.
-        best_epoch = trial.best_step
+        # Reload best checkpoint.
+        # Only load weights to avoid loading `custom_objects`.
         with maybe_distribute(self.distribution_strategy):
-            model.load_weights(
-                self._get_checkpoint_fname(trial.trial_id, best_epoch)
-            )
+            model.load_weights(self._get_checkpoint_fname(trial.trial_id))
         return model
 
     def on_batch_begin(self, trial, model, batch, logs):
@@ -365,8 +328,17 @@ class Tuner(base_tuner.BaseTuner):
         pass
 
     def on_epoch_end(self, trial, model, epoch, logs=None):
+        """Called at the end of an epoch.
+
+        Args:
+            trial: A `Trial` instance.
+            model: A Keras `Model`.
+            epoch: The current epoch number.
+            logs: Dict. Metrics for this epoch. This should include
+              the value of the objective for this epoch.
+        """
         # Intermediate results are not passed to the Oracle, and
-        # checkpointing is handled via a `ModelCheckpoint` callback.
+        # checkpointing is handled via a `SaveBestEpoch` callback.
         pass
 
     def get_best_models(self, num_models=1):
@@ -388,13 +360,13 @@ class Tuner(base_tuner.BaseTuner):
             List of trained model instances sorted from the best to the worst.
         """
         # Method only exists in this class for the docstring override.
-        return super(Tuner, self).get_best_models(num_models)
+        return super().get_best_models(num_models)
 
     def _deepcopy_callbacks(self, callbacks):
         try:
             callbacks = copy.deepcopy(callbacks)
         except:
-            raise ValueError(
+            raise errors.FatalValueError(
                 "All callbacks used during a search "
                 "should be deep-copyable (since they are "
                 "reused across trials). "
@@ -420,29 +392,15 @@ class Tuner(base_tuner.BaseTuner):
                 )
 
     def _get_tensorboard_dir(self, logdir, trial_id, execution):
-        return os.path.join(str(logdir), str(trial_id), "execution" + str(execution))
+        return os.path.join(str(logdir), str(trial_id), f"execution{str(execution)}")
 
-    def _get_checkpoint_dir(self, trial_id, epoch):
-        return os.path.join(
-            self.get_trial_dir(trial_id), "checkpoints", "epoch_" + str(epoch)
-        )
-
-    def _get_checkpoint_fname(self, trial_id, epoch):
+    def _get_checkpoint_fname(self, trial_id):
         return os.path.join(
             # Each checkpoint is saved in its own directory.
-            self._get_checkpoint_dir(trial_id, epoch),
+            self.get_trial_dir(trial_id),
             "checkpoint",
         )
 
-    def _checkpoint_model(self, model, trial_id, epoch):
-        fname = self._get_checkpoint_fname(trial_id, epoch)
-        # Save in TF format.
-        model.save_weights(fname)
-        return fname
-
-    def _delete_checkpoint(self, trial_id, epoch):
-        tf.io.gfile.rmtree(self._get_checkpoint_dir(trial_id, epoch))
-
 
 def maybe_compute_model_size(model):
     """Compute the size of a given model, if it has been built."""
